/* Reading of sequences in FASTA format */
/* 
   Copyright (c) 2004-2005 by Igor B. Kuznetsov. All Rights Reserved.
   IKuznetsov@albany.edu

   Licensor hereby grants to any person a permission to use, copy and distribute any part of this software
   program for educational and non-profit purposes, without fee, and without a written agreement,
   provided that this copyright notice appears in all copies. You cannot modify this source code without 
   an explicit written permission of the author. 

   Any unauthorized commercial distribution of any part of this software program is strictly prohibited. 

   This software is provided on an "AS IS" BASIS and WITHOUT WARRANTY, either express or implied, 
   including, without limitation, the warranties of NON-INFRINGEMENT, MERCHANTABILITY 
   or FITNESS FOR A PARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY OF THE SOFTWARE IS WITH YOU.

   Limitation of Liability. Under no circumstances and under no legal theory shall the Licensor be liable 
   to any person for any direct, indirect, special, incidental, or consequential damages of any character 
   arising as a result of the use of this software. 
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "lib.h"

#define STRING_SIZE 20000

static char _string_[STRING_SIZE];

int NumberOfSequences(FILE *fil)
{
int n=0;
char *c;

for(;;)
{
	c = fgets(_string_,STRING_SIZE,fil);
    if(c == NULL || feof(fil)) {rewind(fil);return(n);}

    if(_string_[strlen(_string_)-1] == '\n') _string_[strlen(_string_)-1] = '\0';
    else while(!feof(fil) && getc(fil) != '\n');

    if(_string_[0] == '>') n++;/*printf("N=%d\n",n);*/
}

}

int ReadSequence(long pos,FILE *fil,char *seq,int len)
{
int i;
char c;

fseek(fil,pos,SEEK_SET);
for(i = 0;i < len;i++)
{
 if(feof(fil)) return(0);
 c = getc(fil);
 if(c == ' ' || c == '\n' || c == '\t') i--;
 else seq[i] = c;
}
seq[i]='\0';
return(1);
}

int ReadFastaFile(struct FASTA_FILE *f)
{
int i,num;

if((num=NumberOfSequences(f->fil)) == 0) return(0);
f->n = num;
if((f->l=(int*)calloc(num+1,sizeof(int))) == NULL ||
	(f->name=(char**)calloc(num+1,sizeof(char*))) == NULL ||
	(f->seq=(char**)calloc(num+1,sizeof(char*))) == NULL) return(0);

for(i = 0;i < num;i++)
	 if(ReadFastaEntry(f,i) == 0) return(0);

return(1);
}


int ReadFastaEntry(struct FASTA_FILE *f,int num)
{
int i,len;
long pos;
char c;

while((c=getc(f->fil)) != '>');

fgets(_string_,STRING_SIZE,f->fil);
if(_string_[strlen(_string_)-1] == '\n') _string_[strlen(_string_)-1] = '\0';
else while(getc(f->fil) != '\n');

if((f->name[num]=(char*)calloc(strlen(_string_)+1,sizeof(char))) == NULL) return(0);
strcpy(f->name[num],_string_);

pos = ftell(f->fil);
for(i = 0;;)
{
  c = getc(f->fil);
  if(isalpha(c) || c == '-') i++;
  if(c == '>' || feof(f->fil)) {ungetc(c,f->fil);break;}
}
f->l[num] = len = i;

if((f->seq[num]=(char*)calloc(len+1,sizeof(char))) == NULL) return(0);
if(ReadSequence(pos,f->fil,f->seq[num],len) == 0) return(0);

return(1);
}

/* Reads one sequence, returns sequence length */
int ReadFastaSequence2(char *seq,FILE *in)
{
int i,len;
long pos;
char c;

while((c=getc(in)) != '>');

fgets(_string_,STRING_SIZE,in);
if(_string_[strlen(_string_)-1] == '\n') _string_[strlen(_string_)-1] = '\0';
else while(getc(in) != '\n');

pos = ftell(in);
for(i = 0;;)
{
  c = getc(in);
  if(isalpha(c) || c == '-') i++;
  if(c == '>' || feof(in)) {ungetc(c,in);break;}
}
len = i;

if(ReadSequence(pos,in,seq,len) == 0) return(0);

return(len);
}

/* Reads one sequence, returns seq. length on success, 0 otherwise */
int ReadFastaSequence(char *seq,int l,FILE *in)
{
int i;
char c;

while((c=getc(in)) != '>');

fgets(_string_,STRING_SIZE,in);
if(_string_[strlen(_string_)-1] == '\n') _string_[strlen(_string_)-1] = '\0';
else while(getc(in) != '\n');

for(i = 0;;)
{
  c = getc(in);
  if(isalpha(c) || c == '-')
  {
      if(++i > l-1) {fprintf(stderr,"\nSequence too long in ReadFastaSequence\n");return(0);};
      seq[i-1] = c;
  }
  if(c == '>' || feof(in)) {ungetc(c,in);seq[i] = '\0';break;}
}

return(i);
}

/* Reads one sequence and its name, returns seq. length on success, 0 otherwise */
int ReadFastaEntry2(char *name,char *seq,int l,FILE *in)
{
int i;
char c;

while((c=getc(in)) != '>');

fgets(_string_,STRING_SIZE,in);
if(_string_[strlen(_string_)-1] == '\n') _string_[strlen(_string_)-1] = '\0';
else while(getc(in) != '\n');

strcpy(name,_string_);

for(i = 0;;)
{
  c = getc(in);
  if(isalpha(c) || c == '-')
  {
      if(++i > l-1) {fprintf(stderr,"\nSequence too long in ReadFastaSequence\n");return(0);};
      seq[i-1] = c;
  }
  if(c == '>' || feof(in)) {ungetc(c,in);seq[i] = '\0';break;}
}

return(i);
}

int ReadLine(char *sss,FILE *fil)
{
int i;
char c;

i = 0;
while((c=getc(fil)) != '\n') if(isalpha(c)) {sss[i] = c;i++;}
sss[i] = '\0';
return(i);
}

/* Find a letter in an array. Return -1 if not found. */
int FindLetter(char c,int alph,char *aa)
{
int i;

for(i = 0;i < alph;i++) if(aa[i] == c) return(i);
return(-1);
}

/* Return number of an amino acid group according to the collapsed alphabet */
int ReturnAAGroup(char c,struct ALPHABET *a)
{
int i,j;

for(i = 0;i < a->number;i++)
for(j = 0;j < a->size[i];j++) if(c == a->group[i][j]) return(i+1);
return(0);
}
